from datetime import datetime, timedelta
from enum import Enum

from pydantic import BaseModel, Field


class PhaseStatus(str, Enum):
    waiting = 'waiting'
    running = 'running'
    closed = 'closed'


class Guess(BaseModel):
    time: datetime
    coordinates: str


class Phase(BaseModel):
    start_time: datetime = Field(default_factory=datetime.now)
    duration: timedelta
    guesses: dict[str, Guess] = {}  # user ID to guess



from pymongo.errors import DuplicateKeyError

from moontour_common.database.clients.mongo import mongo_client
from moontour_common.models.user import InternalUser

users_collection = mongo_client.get_collection('users')


class UserExistsException(Exception):
    pass


def get_user(username: str) -> InternalUser | None:
    user_dict = users_collection.find_one({'username': username})
    return InternalUser(**user_dict) if user_dict else None


def create_user(user: InternalUser):
    try:
        users_collection.insert_one(user.json())
    except DuplicateKeyError:
        raise UserExistsException(f'User "{user.username}" already exists')



from redis.client import Pipeline
from redis.commands.json.path import Path
from redis.commands.search.query import Query

from moontour_common.database.clients.redis import redis_client
from moontour_common.models.phase import Phase
from moontour_common.models.room import Room

ROOMS_INDEX = 'rooms_idx'
ROOM_KEY_PREFIX = 'room:'


class RoomJoiningException(Exception):
    pass


def _get_room_key(room_id: str):
    return f'{ROOM_KEY_PREFIX}{room_id}'


def create_room(room: Room):
    redis_client.json().set(_get_room_key(room.id), Path.root_path(), room.json(exclude={'id'}))


def get_room(room_id: str) -> Room:
    room_dict = redis_client.json().get(_get_room_key(room_id), Path.root_path())
    return Room(**room_dict)


def join_room(room_id: str, user_id: str):
    room_key = _get_room_key(room_id)

    def transaction(pipe: Pipeline):
        if pipe.json().arrlen(room_key, Path('.players')) >= pipe.json().get(room_key, Path('.max_player_count')):
            raise RoomJoiningException('Room is full')

        if pipe.json().get(room_key, Path('.start_time')) is not None:
            raise RoomJoiningException('Room is already running')

        pipe.json().arrappend(room_key, Path('.players'), user_id)

    redis_client.transaction(transaction, room_key)


def find_waiting_room() -> Room | None:
    result = redis_client.ft(ROOMS_INDEX).search(Query('@status:{waiting}'))
    if len(result.docs) == 0:
        return None
    room = Room(**result.docs[0])
    return room


def add_phase(room_id: str, phase: Phase):
    redis_client.json().arrappend(_get_room_key(room_id), Path('.phases'), phase)



import uuid

from pydantic import BaseModel, Field


class UserBase(BaseModel):
    username: str
    full_name: str | None = None


class User(UserBase):
    id: str = Field(alias='id_', default_factory=lambda: str(uuid.uuid4()))


class UserCreate(UserBase):
    password: str


class InternalUser(User):
    hashed_password: str



import uuid
from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field

from moontour_common.models.phase import Phase


class RoomStatus(str, Enum):
    waiting = 'waiting'
    running = 'running'
    closed = 'closed'


class Player(BaseModel):
    user_id: str
    score: int = 0


class Room(BaseModel):
    id: str = Field(alias='id_', default_factory=lambda: str(uuid.uuid4()))
    status: RoomStatus = RoomStatus.waiting
    create_time: datetime = Field(default_factory=datetime.now)
    start_time: datetime | None = None
    phase_count: int = 5
    phases: list[Phase] = []
    max_player_count: int = 4
    players: set[Player] = []
